/* ---------------------------------------------------------------------
	Program to manipulate columns of data in a file.

	Copyright (c) 1987-1994 by CompuSource. All rights reserved worldwide.

	Version	1.2
   --------------------------------------------------------------------- */

/* ---------------------------------------------------------------------
	Include definitions of structures, constants, functions, etc.
   --------------------------------------------------------------------- */

#include	<stdio.h>
#include	<stdlib.h>

#include	"column.hpp"
#include	"fileio.hpp"
#include	"params.hpp"
#include	"transfrm.hpp"

/* ---------------------------------------------------------------------
	Locally used variables.
   --------------------------------------------------------------------- */

PARAMS		params;
TRANSFORM	transform;
FILEIO		inFile;
FILEIO		outFile;

/* ---------------------------------------------------------------------
	Produce the appropriate error message and abort execution.
   --------------------------------------------------------------------- */

const char	*err_msgs[] = {
/*  0 */	"Unknown error",
/*  1 */	"Unable to open file",
/*  2 */	"Could not assign a file buffer for file",
/*  3 */	"Unable to close file",
/*  4 */	"Insufficient memory for",
/*  5 */	"Unknown command line specifier",
/*  6 */	"End point must not be smaller than start",
/*  7 */	"Cannot make limited range changes if record size alters",
/*  8 */	"Overlapping column requirements",
/*  9 */	"Column range outside limits of input file",
/* 10 */	"Unable to process sufficient information for",
/* 11 */	"Error with column permutation discovered during",
		};

void	error_handler(unsigned int errnum, char *message)
{
	fprintf(stderr,"%s: %s\n",
		err_msgs[(errnum < 1 || errnum > 11) ? 0 : errnum],
		(message == 0) ? "" : message);
	fprintf(stderr,"Type COLUMN for help information\n");
	exit(errnum);
}

/* ---------------------------------------------------------------------
	Display help on how to use the program. Exit program afterwards.
   --------------------------------------------------------------------- */

void	help(void)
{
	fprintf(stderr,
		"Usage:\tCOLUMN infile outfile Q Tl/Fl Lr1,r2 [Rc1,c2,Cc/Mc/D/Z/B/Sx]\n"
		"\nwhere\tQ\t= don't execute, but show how transformed\n"
		"\tT/F\t= text file or fixed length data file\n"
		"\tl\t= maximum record length (maximum 32000)\n"
		"\tL\t= specification of which lines to operate on\n"
		"\tr1,r2\t= inclusive range of lines (records)\n"
		"\tR\t= specification of column range and action\n"
		"\tc1,c2\t= inclusive range if columns\n"
		"\tC/M/D\t= whether to copy & insert, move or delete the columns\n"
		"\tc\t= new column location if copy or move\n"
		"\tZ/B/S\t= whether to zero, blank or set the columns\n"
		"\tx\t= new column values i.e. number,number,number,...\n"
		"\nExample\tcolumn in out R1,5,D R6,10,C1\n"
		"\tcolumn in out L20,29 F40 R31,40,M11 R1,3,S1,0,0\n"
		"\nNotes:\tA column range specification must be given.\n"
		"\tMore than one column range may be specified.\n"
		"\tThe input file must be different to the output file.\n"
		"\tDefaults are text mode, l = 255, all records.\n"
		"\tSpecifications may be given in any order.\n"
	);

	exit(99);
}

/* ---------------------------------------------------------------------
	Read the value of a parameter and update the string pointer.
	The arguments are:
		param		=	pointer to char pointer of parameter (modified)
		default	=	default value if parameter non-existent or zero.
	Return value is the parameter (or its default if zero).
   --------------------------------------------------------------------- */

int	read_param(char **param, int defolt)
{
	long	actual;

	actual = strtol(*param, param, 10);
	if (**param != '\0') (*param)++;

	return (actual == 0) ? defolt : actual;
}

/* ---------------------------------------------------------------------
	Process the command line arguments.
	Valid specifications are for:
		text / fixed length files
		record numbers to process
		column ranges and what to do with them.
	No return value.
   --------------------------------------------------------------------- */

void	arguments(int argc, char **argv)
{
	int	iNewRecLen;

	/*	read the main parameters and initialise the transformation tables */

	iNewRecLen = params.Setup(argc, argv);

	transform.Initialise(params.iRecLen, iNewRecLen);

	/*	work out the permutation table and which columns to modify */

	transform.Setup(params.iRecLen, argc, argv);
}

/* ---------------------------------------------------------------------
	Show the user how the file's columns will be transformed.
   --------------------------------------------------------------------- */

void	Transform::show(PARAMS* pParams)
{
	int	i, iMax;

	iMax = iNewRecLen;
	if (pParams->fTextMode) {
		iMax = 0;
		for (i=iNewRecLen-1; i>=0; i--) {
			if (pchSet[i] || (piPermute[i] != i+1)) {
				iMax = i+1;
				break;
			}
		}
	}

	if (iMax == 0) {
		printf("No column translation will occur!\n");
		return;
	}

	printf("New col #   Old col #   Set to\n");
	for (i=0; i<iMax; i++) {
		printf("  %5u     ", i+1);
		if (pchSet[i])
			printf("             %3u\n", pchValues[i]);
		else
			printf("  %5u\n", piPermute[i]);
	}

	if (iMax < iNewRecLen) {
		printf("      :           :\n  %5u       %5u\n", iNewRecLen, iNewRecLen);
	}
}

/* ---------------------------------------------------------------------
	Read and process each record in turn.
   --------------------------------------------------------------------- */

void	process(void)
{
	int	recnum = 0;

	while (inFile.read_record()) {
		recnum++;
		if (params.in_range(recnum)) {
			transform.Record(inFile.pchRecord, outFile.pchRecord, params.fTextMode);
			outFile.write_record(outFile.pchRecord, outFile.iRecLen);
		}
		else {
			outFile.write_record(inFile.pchRecord, inFile.iRecLen);
		}
	}
}

/* ---------------------------------------------------------------------
	Close all files and free dynamically allocated memory.
	The arguments are:
		fFiles	=	says whether to close files
	No return value.
   --------------------------------------------------------------------- */

void	goodbye(BOOL fFiles)
{
	if (fFiles) {
		inFile.Cleanup();
		outFile.Cleanup();
	}

	transform.Cleanup();
}

/* ---------------------------------------------------------------------
	The main entry point of the COLUMN program.
   --------------------------------------------------------------------- */

int	main(int argc, char **argv)
{
	fprintf(stderr,
		"\nCOLUMN manipulation program\n"
		"\nCopyright (c) 1987-1994 by CompuSource. All rights reserved worldwide.\n\n"
	);

	if (argc < 4) help();

	params.defaults();
	arguments(argc, argv);

	if (params.fQuestion) {
		transform.show(&params);
	}
	else {
		inFile.buffers(argv[1], TRUE,		params.iRecLen,			params.fTextMode);
		outFile.buffers(argv[2], FALSE,	transform.iNewRecLen,	params.fTextMode);
		process();
	}

	goodbye(params.fQuestion == 0);

	return 0;
}


